# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from typing import (Optional, List)
from azure.mgmt.cdn.models import (AFDEndpoint, HealthProbeRequestType, EnabledState, Route, LinkToDefaultDomain,
                                   ResourceReference, AFDEndpointProtocols, HttpsRedirect, ForwardingProtocol,
                                   QueryStringCachingBehavior, HealthProbeParameters, MatchProcessingBehavior,
                                   AFDOrigin, AFDOriginGroup, SharedPrivateLinkResourceProperties, CompressionSettings,
                                   LoadBalancingSettingsParameters, SecurityPolicyWebApplicationFirewallParameters,
                                   SecurityPolicyWebApplicationFirewallAssociation, CustomerCertificateParameters,
                                   AFDDomain, AFDDomainHttpsParameters, AfdCertificateType, AfdMinimumTlsVersion,
                                   AFDEndpointUpdateParameters, SkuName, AfdPurgeParameters, Secret,
                                   SecurityPolicy, ProfileUpdateParameters)

from azure.mgmt.cdn.operations import (AFDOriginGroupsOperations, AFDOriginsOperations, AFDProfilesOperations,
                                       SecretsOperations, AFDEndpointsOperations, RoutesOperations, RuleSetsOperations,
                                       RulesOperations, SecurityPoliciesOperations, AFDCustomDomainsOperations,
                                       ProfilesOperations)

from azure.cli.core.commands.client_factory import get_subscription_id
from azure.cli.core.util import (sdk_no_wait)
from azure.cli.core.azclierror import (InvalidArgumentValueError)
from azure.core.exceptions import (ResourceNotFoundError)

from knack.log import get_logger
from msrest.polling import LROPoller, NoPolling

from .custom import _update_mapper

logger = get_logger(__name__)


def default_content_types():
    return ["application/eot",
            "application/font",
            "application/font-sfnt",
            "application/javascript",
            "application/json",
            "application/opentype",
            "application/otf",
            "application/pkcs7-mime",
            "application/truetype",
            "application/ttf",
            "application/vnd.ms-fontobject",
            "application/xhtml+xml",
            "application/xml",
            "application/xml+rss",
            "application/x-font-opentype",
            "application/x-font-truetype",
            "application/x-font-ttf",
            "application/x-httpd-cgi",
            "application/x-javascript",
            "application/x-mpegurl",
            "application/x-opentype",
            "application/x-otf",
            "application/x-perl",
            "application/x-ttf",
            "font/eot",
            "font/ttf",
            "font/otf",
            "font/opentype",
            "image/svg+xml",
            "text/css",
            "text/csv",
            "text/html",
            "text/javascript",
            "text/js",
            "text/plain",
            "text/richtext",
            "text/tab-separated-values",
            "text/xml",
            "text/x-script",
            "text/x-component",
            "text/x-java-source"]


def create_afd_profile(client: ProfilesOperations, resource_group_name, profile_name,
                       sku: SkuName,
                       tags=None):
    from azure.mgmt.cdn.models import (Profile, Sku)

    # Force location to global
    profile = Profile(location="global", sku=Sku(name=sku), tags=tags)
    return client.begin_create(resource_group_name, profile_name, profile)


def delete_afd_profile(client: ProfilesOperations, resource_group_name, profile_name):
    profile = None
    try:
        profile = client.get(resource_group_name, profile_name)
    except ResourceNotFoundError:
        pass

    if profile is None or profile.sku.name not in (SkuName.premium_azure_front_door,
                                                   SkuName.standard_azure_front_door):
        def get_long_running_output(_):
            return None

        logger.warning("Unexpected SKU type, only Standard_AzureFrontDoor and Premium_AzureFrontDoor are supported.")
        return LROPoller(client, None, get_long_running_output, NoPolling())

    return client.begin_delete(resource_group_name, profile_name)


def update_afd_profile(client: ProfilesOperations, resource_group_name, profile_name, tags):
    profile = client.get(resource_group_name, profile_name)
    if profile.sku.name not in (SkuName.premium_azure_front_door, SkuName.standard_azure_front_door):
        logger.warning('Unexpected SKU type, only Standard_AzureFrontDoor and Premium_AzureFrontDoor are supported')
        raise ResourceNotFoundError("Operation returned an invalid status code 'Not Found'")

    return client.begin_update(resource_group_name, profile_name, ProfileUpdateParameters(tags=tags))


def list_afd_profiles(client: ProfilesOperations, resource_group_name=None):
    profile_list = client.list_by_resource_group(resource_group_name=resource_group_name) \
        if resource_group_name else client.list()

    profile_list = [profile for profile in profile_list if profile.sku.name in
                    (SkuName.premium_azure_front_door, SkuName.standard_azure_front_door)]

    return list(profile_list)


def get_afd_profile(client: ProfilesOperations, resource_group_name, profile_name):
    profile = client.get(resource_group_name, profile_name)
    if profile.sku.name not in (SkuName.premium_azure_front_door, SkuName.standard_azure_front_door):
        # Workaround to make the behavior consist with true "Not Found"
        logger.warning('Unexpected SKU type, only Standard_AzureFrontDoor and Premium_AzureFrontDoor are supported')
        raise ResourceNotFoundError("Operation returned an invalid status code 'Not Found'")

    return profile


def list_resource_usage(client: AFDProfilesOperations, resource_group_name, profile_name):
    return client.list_resource_usage(resource_group_name, profile_name)


def create_afd_endpoint(client: AFDEndpointsOperations, resource_group_name, profile_name, endpoint_name,
                        origin_response_timeout_seconds,
                        enabled_state, tags=None, no_wait=None):

    # Force location to global
    endpoint = AFDEndpoint(location="global",
                           origin_response_timeout_seconds=origin_response_timeout_seconds,
                           enabled_state=enabled_state,
                           tags=tags)

    return sdk_no_wait(no_wait, client.begin_create, resource_group_name, profile_name, endpoint_name, endpoint)


def purge_afd_endpoint_content(client: AFDEndpointsOperations, resource_group_name, profile_name, endpoint_name,
                               content_paths, domains=None, no_wait=None):
    endpoint = AfdPurgeParameters(content_paths=content_paths,
                                  domains=domains)

    return sdk_no_wait(no_wait, client.begin_purge_content, resource_group_name, profile_name, endpoint_name, endpoint)


def update_afd_endpoint(client: AFDEndpointsOperations, resource_group_name, profile_name, endpoint_name,
                        origin_response_timeout_seconds=None, enabled_state=None, tags=None):
    update_properties = AFDEndpointUpdateParameters(
        origin_response_timeout_seconds=origin_response_timeout_seconds,
        enabled_state=enabled_state,
        tags=tags
    )

    return client.begin_update(resource_group_name, profile_name, endpoint_name, update_properties)


def create_afd_origin_group(client: AFDOriginGroupsOperations,
                            resource_group_name: str,
                            profile_name: str,
                            origin_group_name: str,
                            load_balancing_sample_size: int,
                            load_balancing_successful_samples_required: int,
                            load_balancing_additional_latency_in_milliseconds: int,
                            probe_request_type: HealthProbeRequestType,
                            probe_protocol: str,
                            probe_path: str,
                            probe_interval_in_seconds: int = 240):

    # Add response error detection support once RP support it.
    health_probe_parameters = HealthProbeParameters(probe_path=probe_path,
                                                    probe_request_type=probe_request_type,
                                                    probe_protocol=probe_protocol,
                                                    probe_interval_in_seconds=probe_interval_in_seconds)

    load_balancing_settings_parameters = LoadBalancingSettingsParameters(
        sample_size=load_balancing_sample_size,
        successful_samples_required=load_balancing_successful_samples_required,
        additional_latency_in_milliseconds=load_balancing_additional_latency_in_milliseconds)

    afd_origin_group = AFDOriginGroup(load_balancing_settings=load_balancing_settings_parameters,
                                      health_probe_settings=health_probe_parameters)

    return client.begin_create(resource_group_name,
                               profile_name,
                               origin_group_name,
                               afd_origin_group).result()


def update_afd_origin_group(client: AFDOriginGroupsOperations,
                            resource_group_name: str,
                            profile_name: str,
                            origin_group_name: str,
                            load_balancing_sample_size: int = None,
                            load_balancing_successful_samples_required: int = None,
                            load_balancing_additional_latency_in_milliseconds: int = None,
                            probe_request_type: HealthProbeRequestType = None,
                            probe_protocol: str = None,
                            probe_path: str = None,
                            probe_interval_in_seconds: int = None):

    # Move these to the parameters list once support is added in RP.
    existing = client.get(resource_group_name, profile_name, origin_group_name)

    health_probe_parameters = HealthProbeParameters(probe_path=probe_path,
                                                    probe_request_type=probe_request_type,
                                                    probe_protocol=probe_protocol,
                                                    probe_interval_in_seconds=probe_interval_in_seconds)

    _update_mapper(existing.health_probe_settings,
                   health_probe_parameters,
                   ["probe_path", "probe_request_type", "probe_protocol", "probe_interval_in_seconds"])

    load_balancing_settings_parameters = LoadBalancingSettingsParameters(
        sample_size=load_balancing_sample_size,
        successful_samples_required=load_balancing_successful_samples_required,
        additional_latency_in_milliseconds=load_balancing_additional_latency_in_milliseconds)
    _update_mapper(existing.load_balancing_settings,
                   load_balancing_settings_parameters,
                   ["sample_size", "successful_samples_required", "additional_latency_in_milliseconds"])

    afd_origin_group = AFDOriginGroup(load_balancing_settings=load_balancing_settings_parameters,
                                      health_probe_settings=health_probe_parameters)

    return client.begin_create(resource_group_name,
                               profile_name,
                               origin_group_name,
                               afd_origin_group).result()


def create_afd_origin(client: AFDOriginsOperations,
                      resource_group_name: str,
                      profile_name: str,
                      origin_group_name: str,
                      origin_name: str,
                      host_name: str,
                      enabled_state: EnabledState,
                      enable_private_link: bool = None,
                      private_link_resource: str = None,
                      private_link_location: str = None,
                      private_link_sub_resource_type: str = None,
                      private_link_request_message: str = None,
                      http_port: int = 80,
                      https_port: int = 443,
                      origin_host_header: Optional[str] = None,
                      priority: int = 1,
                      weight: int = 1000):

    shared_private_link_resource = None
    if enable_private_link:
        shared_private_link_resource = SharedPrivateLinkResourceProperties(
            private_link=ResourceReference(id=private_link_resource),
            private_link_location=private_link_location,
            group_id=private_link_sub_resource_type,
            request_message=private_link_request_message)

    return client.begin_create(resource_group_name,
                               profile_name,
                               origin_group_name,
                               origin_name,
                               AFDOrigin(
                                   host_name=host_name,
                                   http_port=http_port,
                                   https_port=https_port,
                                   origin_host_header=origin_host_header,
                                   priority=priority,
                                   weight=weight,
                                   shared_private_link_resource=shared_private_link_resource,
                                   enabled_state=enabled_state))


def update_afd_origin(client: AFDOriginsOperations,
                      resource_group_name: str,
                      profile_name: str,
                      origin_group_name: str,
                      origin_name: str,
                      host_name: str = None,
                      enabled_state: EnabledState = None,
                      http_port: int = None,
                      https_port: int = None,
                      origin_host_header: Optional[str] = None,
                      priority: int = None,
                      weight: int = None,
                      enable_private_link: bool = None,
                      private_link_resource: str = None,
                      private_link_location: str = None,
                      private_link_sub_resource_type: str = None,
                      private_link_request_message: str = None):

    existing = client.get(resource_group_name, profile_name, origin_group_name, origin_name)
    origin = AFDOrigin(
        host_name=host_name,
        http_port=http_port,
        https_port=https_port,
        origin_host_header=origin_host_header,
        priority=priority,
        weight=weight,
        enabled_state=enabled_state)

    _update_mapper(
        existing,
        origin,
        ["host_name", "http_port", "https_port", "origin_host_header", "priority", "weight", "enabled_state"])

    if enable_private_link is not None and not enable_private_link:
        origin.shared_private_link_resource = None
    elif (private_link_resource is not None or
          private_link_location is not None or
          private_link_sub_resource_type is not None or
          private_link_request_message is not None):
        shared_private_link_resource = SharedPrivateLinkResourceProperties(
            private_link=ResourceReference(id=private_link_resource) if private_link_resource is not None else None,
            private_link_location=private_link_location,
            group_id=private_link_sub_resource_type,
            request_message=private_link_request_message)

        if existing.shared_private_link_resource is not None:
            existing_shared_private_link_resource = SharedPrivateLinkResourceProperties(
                private_link=ResourceReference(id=existing.shared_private_link_resource["privateLink"]['id']),
                private_link_location=existing.shared_private_link_resource["privateLinkLocation"],
                group_id=existing.shared_private_link_resource["groupId"],
                request_message=existing.shared_private_link_resource["requestMessage"])

            _update_mapper(
                existing_shared_private_link_resource,
                shared_private_link_resource,
                ["private_link", "private_link_location", "group_id", "request_message"])

        origin.shared_private_link_resource = shared_private_link_resource
    else:
        origin.shared_private_link_resource = existing.shared_private_link_resource

    # client.update does not allow unset field
    return client.begin_create(resource_group_name,
                               profile_name,
                               origin_group_name,
                               origin_name,
                               origin)


def create_afd_route(cmd,
                     client: RoutesOperations,
                     resource_group_name: str,
                     profile_name: str,
                     endpoint_name: str,
                     route_name: str,
                     https_redirect: HttpsRedirect,
                     supported_protocols: List[AFDEndpointProtocols],
                     origin_group: str,
                     forwarding_protocol: ForwardingProtocol,
                     link_to_default_domain: LinkToDefaultDomain = None,
                     is_compression_enabled: bool = False,
                     content_types_to_compress: List[str] = None,
                     query_string_caching_behavior: QueryStringCachingBehavior = None,
                     custom_domains: List[str] = None,
                     origin_path: Optional[str] = None,
                     patterns_to_match: List[str] = None,
                     rule_sets: List[str] = None):

    subscription_id = get_subscription_id(cmd.cli_ctx)
    formatted_custom_domains = []
    if custom_domains is not None:
        for custom_domain in custom_domains:
            if '/customdomains/' not in custom_domain.lower():
                custom_domain = f'/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}' \
                                f'/providers/Microsoft.Cdn/profiles/{profile_name}/customDomains/{custom_domain}'

            # If the origin is not an ID, assume it's a name and format it as an ID.
            formatted_custom_domains.append(ResourceReference(id=custom_domain))

    if '/origingroups/' not in origin_group.lower():
        origin_group = f'/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}' \
                       f'/providers/Microsoft.Cdn/profiles/{profile_name}/originGroups/{origin_group}'

    compression_settings = CompressionSettings(
        content_types_to_compress=content_types_to_compress,
        is_compression_enabled=is_compression_enabled
    )

    if is_compression_enabled and content_types_to_compress is None:
        compression_settings.content_types_to_compress = default_content_types()

    if not compression_settings.is_compression_enabled:
        compression_settings.content_types_to_compress = []

    formatted_rule_sets = []
    if rule_sets is not None:
        for rule_set in rule_sets:
            if '/rulesets/' not in rule_set.lower():
                rule_set = f'/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}' \
                           f'/providers/Microsoft.Cdn/profiles/{profile_name}/ruleSets/{rule_set}'

            formatted_rule_sets.append(ResourceReference(id=rule_set))

    return client.begin_create(resource_group_name,
                               profile_name,
                               endpoint_name,
                               route_name,
                               Route(
                                   custom_domains=formatted_custom_domains,
                                   origin_path=origin_path,
                                   patterns_to_match=patterns_to_match if patterns_to_match is not None else ['/*'],
                                   supported_protocols=supported_protocols,
                                   https_redirect=https_redirect,
                                   origin_group=ResourceReference(id=origin_group),
                                   forwarding_protocol=forwarding_protocol,
                                   rule_sets=formatted_rule_sets,
                                   query_string_caching_behavior=query_string_caching_behavior,
                                   compression_settings=compression_settings,
                                   link_to_default_domain=LinkToDefaultDomain.enabled if link_to_default_domain else
                                   LinkToDefaultDomain.disabled))


def update_afd_route(cmd,
                     client: RoutesOperations,
                     resource_group_name: str,
                     profile_name: str,
                     endpoint_name: str,
                     route_name: str,
                     https_redirect: HttpsRedirect = None,
                     supported_protocols: List[AFDEndpointProtocols] = None,
                     origin_group: str = None,
                     forwarding_protocol: ForwardingProtocol = None,
                     link_to_default_domain: LinkToDefaultDomain = None,
                     is_compression_enabled: bool = None,
                     content_types_to_compress: List[str] = None,
                     query_string_caching_behavior: QueryStringCachingBehavior = None,
                     custom_domains: List[str] = None,
                     origin_path: Optional[str] = None,
                     patterns_to_match: List[str] = None,
                     rule_sets: List[str] = None):

    existing = client.get(resource_group_name, profile_name, endpoint_name, route_name)
    route = Route(
        origin_path=origin_path,
        origin_group=existing.origin_group,
        patterns_to_match=patterns_to_match,
        supported_protocols=supported_protocols,
        https_redirect=https_redirect,
        forwarding_protocol=forwarding_protocol,
        query_string_caching_behavior=query_string_caching_behavior,
        link_to_default_domain=link_to_default_domain)

    subscription_id = get_subscription_id(cmd.cli_ctx)
    if custom_domains is not None:
        formatted_custom_domains = []
        for custom_domain in custom_domains:
            if '/customdomains/' not in custom_domain.lower():
                custom_domain = f'/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}' \
                                f'/providers/Microsoft.Cdn/profiles/{profile_name}/customDomains/{custom_domain}'

            # If the origin is not an ID, assume it's a name and format it as an ID.
            formatted_custom_domains.append(ResourceReference(id=custom_domain))

        route.custom_domains = formatted_custom_domains

    if origin_group is not None:
        if '/origingroups/' not in origin_group.lower():
            origin_group = f'/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}' \
                           f'/providers/Microsoft.Cdn/profiles/{profile_name}/originGroups/{origin_group}'

        route.origin_group = origin_group

    if rule_sets is not None:
        formatted_rule_sets = []
        for rule_set in rule_sets:
            if '/rulesets/' not in rule_set.lower():
                rule_set = f'/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}' \
                           f'/providers/Microsoft.Cdn/profiles/{profile_name}/ruleSets/{rule_set}'

            # If the origin is not an ID, assume it's a name and format it as an ID.
            formatted_rule_sets.append(ResourceReference(id=rule_set))

        route.rule_sets = formatted_rule_sets

    _update_mapper(existing, route,
                   ["custom_domains", "origin_path", "patterns_to_match",
                    "supported_protocols", "https_redirect", "origin_group",
                    "forwarding_protocol", "rule_sets", "query_string_caching_behavior",
                    "link_to_default_domain", "compression_settings"])

    if is_compression_enabled:
        if existing.compression_settings is None:
            route.compression_settings = CompressionSettings(
                content_types_to_compress=content_types_to_compress if content_types_to_compress is not None else
                default_content_types(),
                is_compression_enabled=is_compression_enabled
            )
        else:
            route.compression_settings = CompressionSettings(
                content_types_to_compress=content_types_to_compress if content_types_to_compress is not None else
                existing.compression_settings.content_types_to_compress,
                is_compression_enabled=is_compression_enabled
            )
    elif is_compression_enabled is None:
        if content_types_to_compress is not None and existing.compression_settings is not None:
            route.compression_settings = CompressionSettings(
                content_types_to_compress=content_types_to_compress,
                is_compression_enabled=existing.compression_settings["isCompressionEnabled"]
            )
    else:
        route.compression_settings = CompressionSettings(
            content_types_to_compress=[],
            is_compression_enabled=False
        )

    return client.begin_create(resource_group_name,
                               profile_name,
                               endpoint_name,
                               route_name,
                               route)


def create_afd_rule_set(client: RuleSetsOperations,
                        resource_group_name: str,
                        profile_name: str,
                        rule_set_name: str):

    return client.begin_create(resource_group_name, profile_name, rule_set_name)


# pylint: disable=too-many-locals
def create_afd_rule(client: RulesOperations, resource_group_name, profile_name, rule_set_name,
                    order, rule_name, action_name, match_variable=None, operator=None,
                    match_values=None, selector=None, negate_condition=None, transform=None,
                    cache_behavior=None, cache_duration=None, header_action=None,
                    header_name=None, header_value=None, query_string_behavior=None, query_parameters=None,
                    redirect_type=None, redirect_protocol=None, custom_hostname=None, custom_path=None,
                    custom_querystring=None, custom_fragment=None, source_pattern=None,
                    destination=None, preserve_unmatched_path=None,
                    match_processing_behavior: MatchProcessingBehavior = None):
    from azure.mgmt.cdn.models import Rule
    from .custom import create_condition
    from .custom import create_action

    conditions = []
    condition = create_condition(match_variable, operator, match_values, selector, negate_condition, transform)
    if condition is not None:
        conditions.append(condition)

    actions = []
    action = create_action(action_name, cache_behavior, cache_duration, header_action, header_name,
                           header_value, query_string_behavior, query_parameters, redirect_type,
                           redirect_protocol, custom_hostname, custom_path, custom_querystring,
                           custom_fragment, source_pattern, destination, preserve_unmatched_path)
    if action is not None:
        actions.append(action)

    rule = Rule(
        name=rule_name,
        order=order,
        conditions=conditions,
        actions=actions,
        match_processing_behavior=match_processing_behavior
    )

    return client.begin_create(resource_group_name,
                               profile_name,
                               rule_set_name,
                               rule_name,
                               rule=rule)


def add_afd_rule_condition(client: RulesOperations, resource_group_name, profile_name, rule_set_name,
                           rule_name, match_variable, operator, match_values=None, selector=None,
                           negate_condition=None, transform=None):
    from .custom import create_condition

    existing_rule = client.get(resource_group_name, profile_name, rule_set_name, rule_name)
    condition = create_condition(match_variable, operator, match_values, selector, negate_condition, transform)
    existing_rule.conditions.append(condition)

    return client.begin_create(resource_group_name,
                               profile_name,
                               rule_set_name,
                               rule_name,
                               rule=existing_rule)


def add_afd_rule_action(client: RulesOperations, resource_group_name, profile_name, rule_set_name,
                        rule_name, action_name, cache_behavior=None, cache_duration=None,
                        header_action=None, header_name=None, header_value=None, query_string_behavior=None,
                        query_parameters=None, redirect_type=None, redirect_protocol=None, custom_hostname=None,
                        custom_path=None, custom_querystring=None, custom_fragment=None, source_pattern=None,
                        destination=None, preserve_unmatched_path=None):
    from .custom import create_action

    existing_rule = client.get(resource_group_name, profile_name, rule_set_name, rule_name)
    action = create_action(action_name, cache_behavior, cache_duration, header_action, header_name,
                           header_value, query_string_behavior, query_parameters, redirect_type,
                           redirect_protocol, custom_hostname, custom_path, custom_querystring,
                           custom_fragment, source_pattern, destination, preserve_unmatched_path)

    existing_rule.actions.append(action)
    return client.begin_create(resource_group_name,
                               profile_name,
                               rule_set_name,
                               rule_name,
                               rule=existing_rule)


def remove_afd_rule_condition(client: RulesOperations, resource_group_name, profile_name,
                              rule_set_name, rule_name, index):
    existing_rule = client.get(resource_group_name, profile_name, rule_set_name, rule_name)
    if len(existing_rule.conditions) > 1 and index < len(existing_rule.conditions):
        existing_rule.conditions.pop(index)
    else:
        logger.warning("Invalid condition index found. This command will be skipped. Please check the rule.")

    return client.begin_create(resource_group_name,
                               profile_name,
                               rule_set_name,
                               rule_name,
                               rule=existing_rule)


def remove_afd_rule_action(client: RulesOperations, resource_group_name, profile_name, rule_set_name, rule_name, index):
    existing_rule = client.get(resource_group_name, profile_name, rule_set_name, rule_name)
    if len(existing_rule.actions) > 1 and index < len(existing_rule.actions):
        existing_rule.actions.pop(index)
    else:
        logger.warning("Invalid condition index found. This command will be skipped. Please check the rule.")

    return client.begin_create(resource_group_name,
                               profile_name,
                               rule_set_name,
                               rule_name,
                               rule=existing_rule)


def list_afd_rule_condition(client: RulesOperations, resource_group_name,
                            profile_name, rule_set_name,
                            rule_name):
    rule = client.get(resource_group_name, profile_name, rule_set_name, rule_name)
    return rule.conditions


def list_afd_rule_action(client: RulesOperations, resource_group_name,
                         profile_name, rule_set_name,
                         rule_name):
    rule = client.get(resource_group_name, profile_name, rule_set_name, rule_name)
    return rule.actions


def create_afd_security_policy(client: SecurityPoliciesOperations,
                               resource_group_name,
                               profile_name,
                               security_policy_name,
                               domains: List[str],
                               waf_policy: str):

    if any(("/afdendpoints/" not in domain.lower() and
            "/customdomains/" not in domain.lower()) for domain in domains):
        raise InvalidArgumentValueError('Domain should either be endpoint ID or custom domain ID.')

    if "/frontdoorwebapplicationfirewallpolicies/" not in waf_policy.lower():
        raise InvalidArgumentValueError('waf_policy should be valid Front Door WAF policy ID.')

    # Add patterns and multiple domains support in the feature
    parameters = SecurityPolicyWebApplicationFirewallParameters(
        waf_policy=ResourceReference(id=waf_policy),
        associations=[SecurityPolicyWebApplicationFirewallAssociation(
            domains=[ResourceReference(id=domain) for domain in domains],
            patterns_to_match=["/*"])])

    return client.begin_create(resource_group_name,
                               profile_name,
                               security_policy_name,
                               SecurityPolicy(parameters=parameters))


def update_afd_security_policy(client: SecurityPoliciesOperations,
                               resource_group_name,
                               profile_name,
                               security_policy_name,
                               domains: List[str] = None,
                               waf_policy: str = None):

    if domains is not None and any(("/afdendpoints/" not in domain.lower() and
                                    "/customdomains/" not in domain.lower()) for domain in domains):
        raise InvalidArgumentValueError('Domain should be either endpoint ID or custom domain ID.')

    if waf_policy is not None and "/frontdoorwebapplicationfirewallpolicies/" not in waf_policy:
        raise InvalidArgumentValueError('waf_policy should be Front Door WAF policy ID.')

    existing = client.get(resource_group_name, profile_name, security_policy_name)

    # Add patterns and multiple domains support in the future
    parameters = SecurityPolicyWebApplicationFirewallParameters(
        waf_policy=ResourceReference(id=waf_policy) if waf_policy is not None else existing.parameters.waf_policy,
        associations=[SecurityPolicyWebApplicationFirewallAssociation(
            domains=[ResourceReference(id=domain) for domain in domains],
            patterns_to_match=["/*"])] if domains is not None else existing.parameters.associations)

    return client.begin_create(resource_group_name,
                               profile_name,
                               security_policy_name,
                               SecurityPolicy(parameters=parameters))


def create_afd_secret(client: SecretsOperations,
                      resource_group_name,
                      profile_name,
                      secret_name,
                      secret_source,
                      secret_version: str = None,
                      use_latest_version: bool = None):

    if "/certificates/" not in secret_source.lower():
        raise InvalidArgumentValueError('secret_source should be valid Azure key vault certificate ID.')

    if secret_version is None and not use_latest_version:
        raise InvalidArgumentValueError('Either specify secret_version or enable use_latest_version.')

    # Only support CustomerCertificate for the moment
    parameters = None
    if use_latest_version:
        parameters = CustomerCertificateParameters(
            secret_source=ResourceReference(id=secret_source),
            secret_version=None,
            use_latest_version=True
        )
    else:
        parameters = CustomerCertificateParameters(
            secret_source=ResourceReference(id=f'{secret_source}/{secret_version}'),
            secret_version=secret_version,
            use_latest_version=False
        )

    return client.begin_create(resource_group_name,
                               profile_name,
                               secret_name,
                               Secret(parameters=parameters))


def update_afd_secret(client: SecretsOperations,
                      resource_group_name,
                      profile_name,
                      secret_name,
                      secret_source: str = None,
                      secret_version: str = None,
                      use_latest_version: bool = None):
    existing = client.get(resource_group_name, profile_name, secret_name)

    secret_source_id = secret_source if secret_source is not None else existing.parameters.secret_source.id
    if existing.parameters.secret_version is not None and existing.parameters.secret_version in secret_source_id:
        version_start = secret_source_id.lower().rindex(f'/{existing.parameters.secret_version}')
        secret_source_id = secret_source_id[0:version_start]

    secret_version = (secret_version if secret_version is not None
                      else existing.parameters.secret_version)

    use_latest_version = (use_latest_version if use_latest_version is not None
                          else existing.parameters.use_latest_version)

    return create_afd_secret(client,
                             resource_group_name,
                             profile_name,
                             secret_name,
                             secret_source_id,
                             secret_version,
                             use_latest_version)


def create_afd_custom_domain(cmd,
                             client: AFDCustomDomainsOperations,
                             resource_group_name: str,
                             profile_name: str,
                             custom_domain_name: str,
                             host_name: str,
                             certificate_type: AfdCertificateType,
                             minimum_tls_version: AfdMinimumTlsVersion,
                             azure_dns_zone: str = None,
                             secret: str = None,
                             no_wait: bool = None):

    if azure_dns_zone is not None and "/dnszones/" not in azure_dns_zone.lower():
        raise InvalidArgumentValueError('azure_dns_zone should be valid Azure dns zone ID.')

    subscription_id = get_subscription_id(cmd.cli_ctx)
    if secret is not None and "/secrets/" not in secret.lower():
        secret = f'/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}' \
                 f'/providers/Microsoft.Cdn/profiles/{profile_name}/secrets/{secret}'

    tls_settings = AFDDomainHttpsParameters(certificate_type=certificate_type,
                                            minimum_tls_version=minimum_tls_version,
                                            secret=ResourceReference(id=secret) if secret is not None else None)

    afd_domain = AFDDomain(host_name=host_name,
                           tls_settings=tls_settings,
                           azure_dns_zone=ResourceReference(id=azure_dns_zone) if azure_dns_zone is not None else None)

    return sdk_no_wait(no_wait, client.begin_create, resource_group_name, profile_name, custom_domain_name, afd_domain)


def update_afd_custom_domain(cmd,
                             client: AFDCustomDomainsOperations,
                             resource_group_name: str,
                             profile_name: str,
                             custom_domain_name: str,
                             certificate_type: AfdCertificateType = None,
                             minimum_tls_version: AfdMinimumTlsVersion = None,
                             azure_dns_zone: str = None,
                             secret: str = None):

    if azure_dns_zone is not None and "/dnszones/" not in azure_dns_zone.lower():
        raise InvalidArgumentValueError('azure_dns_zone should be valid Azure dns zone ID.')

    subscription_id = get_subscription_id(cmd.cli_ctx)
    if secret is not None and "/secrets/" not in secret.lower():
        secret = f'/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}' \
                 f'/providers/Microsoft.Cdn/profiles/{profile_name}/secrets/{secret}'

    existing = client.get(resource_group_name, profile_name, custom_domain_name)

    tls_settings = AFDDomainHttpsParameters(certificate_type=certificate_type,
                                            minimum_tls_version=minimum_tls_version,
                                            secret=ResourceReference(id=secret) if secret is not None else None)

    _update_mapper(existing.tls_settings, tls_settings, ["certificate_type", "minimum_tls_version", "secret"])

    if certificate_type == AfdCertificateType.managed_certificate:
        tls_settings.secret = None

    afd_domain = AFDDomain(
        host_name=existing.host_name,
        tls_settings=tls_settings,
        azure_dns_zone=ResourceReference(id=azure_dns_zone) if azure_dns_zone is not None else existing.azure_dns_zone)

    return client.begin_create(resource_group_name,
                               profile_name,
                               custom_domain_name,
                               afd_domain).result()

# endregion



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


def load_arguments(self, _):
    with self.argument_context('find') as c:
        c.positional('cli_term', help='An Azure CLI command or group for which you need an example.')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


def create_log_analytics_workspace_linked_storage_accounts(client, resource_group_name, workspace_name,
                                                           data_source_type, storage_account_ids):
    from azure.mgmt.loganalytics.models import LinkedStorageAccountsResource
    linked_storage_accounts = LinkedStorageAccountsResource(storage_account_ids=storage_account_ids)
    return client.create_or_update(resource_group_name=resource_group_name,
                                   workspace_name=workspace_name,
                                   data_source_type=data_source_type,
                                   parameters=linked_storage_accounts)


def add_log_analytics_workspace_linked_storage_accounts(client, resource_group_name, workspace_name,
                                                        data_source_type, storage_account_ids):
    linked_storage_accounts = client.get(resource_group_name=resource_group_name,
                                         workspace_name=workspace_name,
                                         data_source_type=data_source_type)
    linked_storage_accounts.storage_account_ids.extend(storage_account_ids)
    return client.create_or_update(resource_group_name=resource_group_name,
                                   workspace_name=workspace_name,
                                   data_source_type=data_source_type,
                                   parameters=linked_storage_accounts)


def remove_log_analytics_workspace_linked_storage_accounts(client, resource_group_name, workspace_name,
                                                           data_source_type, storage_account_ids):
    linked_storage_accounts = client.get(resource_group_name=resource_group_name,
                                         workspace_name=workspace_name,
                                         data_source_type=data_source_type)
    storage_account_ids_set = set(str.lower(storage_account_id) for storage_account_id in storage_account_ids)
    for existed_storage_account_id in linked_storage_accounts.storage_account_ids:
        if str.lower(existed_storage_account_id) in storage_account_ids_set:
            linked_storage_accounts.storage_account_ids.remove(existed_storage_account_id)
    return client.create_or_update(resource_group_name=resource_group_name,
                                   workspace_name=workspace_name,
                                   data_source_type=data_source_type,
                                   parameters=linked_storage_accounts)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long

from collections import OrderedDict

from knack.log import get_logger

from azure.cli.core.util import empty_on_404
from azure.cli.core.profiles import ResourceType, PROFILE_TYPE
from azure.cli.core.commands import CliCommandType, DeploymentOutputLongRunningOperation
from azure.cli.core.commands.arm import handle_template_based_exception
from azure.cli.command_modules.resource._client_factory import (
    cf_resource_groups, cf_providers, cf_features, cf_tags, cf_deployments,
    cf_deployment_operations, cf_policy_definitions, cf_policy_set_definitions, cf_policy_exemptions, cf_resource_links,
    cf_resource_deploymentscripts, cf_resource_managedapplications, cf_resource_managedappdefinitions, cf_management_groups, cf_management_group_subscriptions, cf_resource_templatespecs)
from azure.cli.command_modules.resource._validators import process_deployment_create_namespace, process_ts_create_or_update_namespace, _validate_template_spec, _validate_template_spec_out

from ._exception_handler import managementgroups_exception_handler


logger = get_logger(__name__)


# Resource group commands
def transform_resource_group_list(result):
    return [OrderedDict([
        ('Name', r['name']), ('Location', r['location']), ('Status', r['properties']['provisioningState'])]) for r in result]


def transform_resource_list(result):
    transformed = []
    for r in result:
        res = OrderedDict([('Name', r['name']), ('ResourceGroup', r['resourceGroup']), ('Location', r['location']), ('Type', r['type'])])
        try:
            res['Status'] = r['properties']['provisioningStatus']
        except TypeError:
            res['Status'] = ' '
        transformed.append(res)
    return transformed


# Resource group deployment commands
def transform_deployment(result):
    r = result
    return OrderedDict([('Name', r['name']),
                        ('ResourceGroup', r['resourceGroup']),
                        ('State', r['properties']['provisioningState']),
                        ('Timestamp', r['properties']['timestamp']),
                        ('Mode', r['properties']['mode'])])


def transform_deployments_list(result):
    sort_list = sorted(result, key=lambda deployment: deployment['properties']['timestamp'])
    return [transform_deployment(r) for r in sort_list]


# pylint: disable=too-many-statements
def load_command_table(self, _):
    from azure.cli.core.commands.arm import deployment_validate_table_format

    resource_custom = CliCommandType(operations_tmpl='azure.cli.command_modules.resource.custom#{}')

    resource_group_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.resources.operations#ResourceGroupsOperations.{}',
        client_factory=cf_resource_groups,
        resource_type=ResourceType.MGMT_RESOURCE_RESOURCES
    )

    resource_provider_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.resources.operations#ProvidersOperations.{}',
        client_factory=cf_providers,
        resource_type=ResourceType.MGMT_RESOURCE_RESOURCES
    )

    resource_feature_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.features.operations#FeaturesOperations.{}',
        client_factory=cf_features,
        resource_type=ResourceType.MGMT_RESOURCE_FEATURES
    )

    resource_tag_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.resources.operations#TagsOperations.{}',
        client_factory=cf_tags,
        resource_type=ResourceType.MGMT_RESOURCE_RESOURCES
    )

    resource_deployment_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.resources.operations#DeploymentsOperations.{}',
        client_factory=cf_deployments,
        resource_type=ResourceType.MGMT_RESOURCE_RESOURCES
    )

    resource_deployment_operation_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.resources.operations#DeploymentOperationsOperations.{}',
        client_factory=cf_deployment_operations,
        resource_type=ResourceType.MGMT_RESOURCE_RESOURCES
    )

    resource_policy_definitions_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.policy.operations#PolicyDefinitionsOperations.{}',
        client_factory=cf_policy_definitions,
        resource_type=ResourceType.MGMT_RESOURCE_POLICY
    )

    resource_policy_set_definitions_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.policy.operations#PolicySetDefinitionsOperations.{}',
        client_factory=cf_policy_set_definitions,
        resource_type=ResourceType.MGMT_RESOURCE_POLICY
    )

    resource_policy_exemptions_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.policy.operations#PolicyExemptionsOperations.{}',
        client_factory=cf_policy_exemptions,
        resource_type=ResourceType.MGMT_RESOURCE_POLICY
    )

    resource_lock_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.locks.operations#ManagementLocksOperations.{}',
        resource_type=ResourceType.MGMT_RESOURCE_LOCKS
    )

    resource_link_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.links.operations#ResourceLinksOperations.{}',
        client_factory=cf_resource_links,
        resource_type=ResourceType.MGMT_RESOURCE_LINKS
    )

    resource_deploymentscripts_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.deploymentscripts.operations#ResourceLinksOperations.{}',
        client_factory=cf_resource_deploymentscripts,
        resource_type=ResourceType.MGMT_RESOURCE_DEPLOYMENTSCRIPTS
    )

    resource_managedapp_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.managedapplications.operations#ApplicationsOperations.{}',
        client_factory=cf_resource_managedapplications,
        resource_type=ResourceType.MGMT_RESOURCE_RESOURCES
    )

    resource_managedapp_def_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.managedapplications.operations#ApplicationDefinitionsOperations.{}',
        client_factory=cf_resource_managedappdefinitions,
        resource_type=ResourceType.MGMT_RESOURCE_RESOURCES
    )

    resource_managementgroups_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.managementgroups.operations#ManagementGroupsOperations.{}',
        client_factory=cf_management_groups,
        exception_handler=managementgroups_exception_handler
    )

    resource_managementgroups_subscriptions_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.managementgroups.operations#ManagementGroupSubscriptionsOperations.{}',
        client_factory=cf_management_group_subscriptions,
        exception_handler=managementgroups_exception_handler
    )

    resource_managementgroups_update_type = CliCommandType(
        operations_tmpl='azure.cli.command_modules.resource.custom#{}',
        client_factory=cf_management_groups,
        exception_handler=managementgroups_exception_handler
    )

    resource_templatespecs_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.resource.templatespecs.operations#ResourceLinksOperations.{}',
        client_factory=cf_resource_templatespecs,
        resource_type=ResourceType.MGMT_RESOURCE_TEMPLATESPECS
    )

    with self.command_group('account lock', resource_lock_sdk, resource_type=ResourceType.MGMT_RESOURCE_LOCKS) as g:
        g.custom_command('create', 'create_lock')
        g.custom_command('delete', 'delete_lock')
        g.custom_command('list', 'list_locks')
        g.custom_show_command('show', 'get_lock')
        g.custom_command('update', 'update_lock')

    with self.command_group('group', resource_group_sdk, resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.command('delete', 'begin_delete', supports_no_wait=True, confirmation=True)
        g.show_command('show', 'get')
        g.command('exists', 'check_existence')
        g.custom_command('list', 'list_resource_groups', table_transformer=transform_resource_group_list)
        g.custom_command('create', 'create_resource_group')
        g.custom_command('export', 'export_group_as_template')
        g.generic_update_command('update', custom_func_name='update_resource_group', custom_func_type=resource_custom)
        g.wait_command('wait')

    with self.command_group('group lock', resource_type=ResourceType.MGMT_RESOURCE_LOCKS) as g:
        g.custom_command('create', 'create_lock')
        g.custom_command('delete', 'delete_lock')
        g.custom_command('list', 'list_locks')
        g.custom_show_command('show', 'get_lock')
        g.custom_command('update', 'update_lock')

    with self.command_group('resource', resource_custom, resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('create', 'create_resource')
        g.custom_command('delete', 'delete_resource')
        g.custom_show_command('show', 'show_resource')
        g.custom_command('list', 'list_resources', table_transformer=transform_resource_list)
        g.custom_command('tag', 'tag_resource')
        g.custom_command('move', 'move_resource')
        g.custom_command('invoke-action', 'invoke_resource_action', transform=DeploymentOutputLongRunningOperation(self.cli_ctx))
        g.generic_update_command('update', getter_name='show_resource', setter_name='update_resource',
                                 client_factory=None)
        g.wait_command('wait', getter_name='show_resource')

    with self.command_group('resource lock', resource_type=ResourceType.MGMT_RESOURCE_LOCKS) as g:
        g.custom_command('create', 'create_lock')
        g.custom_command('delete', 'delete_lock')
        g.custom_command('list', 'list_locks')
        g.custom_show_command('show', 'get_lock')
        g.custom_command('update', 'update_lock')

    # Resource provider commands
    with self.command_group('provider', resource_provider_sdk, resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.command('list', 'list')
        g.show_command('show', 'get')
        g.custom_command('register', 'register_provider')
        g.custom_command('unregister', 'unregister_provider')
        g.custom_command('operation list', 'list_provider_operations')
        g.custom_command('permission list', 'list_provider_permissions')
        g.custom_show_command('operation show', 'show_provider_operations')

    # Resource feature commands
    with self.command_group('feature', resource_feature_sdk, client_factory=cf_features, resource_type=PROFILE_TYPE,
                            min_api='2019-03-02-hybrid') as g:
        feature_table_transform = '{Name:name, RegistrationState:properties.state}'
        g.custom_command('list', 'list_features', table_transformer='[].' + feature_table_transform)
        g.show_command('show', 'get', table_transformer=feature_table_transform)
        g.custom_command('register', 'register_feature')
        g.custom_command('unregister', 'unregister_feature')

    # Tag commands
    with self.command_group('tag', resource_tag_sdk, resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('list', 'get_tag_at_scope')
        g.custom_command('create', 'create_or_update_tag_at_scope')
        g.custom_command('delete', 'delete_tag_at_scope', confirmation=True)
        g.custom_command('update', 'update_tag_at_scope', min_api='2019-10-01')
        g.command('add-value', 'create_or_update_value')
        g.command('remove-value', 'delete_value')

    # az group deployment
    with self.command_group('group deployment', resource_deployment_sdk, deprecate_info=self.deprecate(redirect='deployment group', hide=True)) as g:
        g.custom_command('create', 'deploy_arm_template', supports_no_wait=True, validator=process_deployment_create_namespace,
                         table_transformer=transform_deployment, exception_handler=handle_template_based_exception)
        g.command('list', 'list_by_resource_group', table_transformer=transform_deployments_list, min_api='2017-05-10')
        g.command('list', 'list', table_transformer=transform_deployments_list, max_api='2016-09-01')
        g.show_command('show', 'get', table_transformer=transform_deployment)
        g.command('delete', 'begin_delete', supports_no_wait=True)
        g.custom_command('validate', 'validate_arm_template', table_transformer=deployment_validate_table_format, exception_handler=handle_template_based_exception)
        g.custom_command('export', 'export_deployment_as_template')
        g.wait_command('wait')
        g.command('cancel', 'cancel')

    with self.command_group('group deployment operation', resource_deployment_operation_sdk, deprecate_info=self.deprecate(redirect='deployment operation group', hide=True)) as g:
        g.command('list', 'list')
        g.custom_show_command('show', 'get_deployment_operations', client_factory=cf_deployment_operations)

    # az deployment
    with self.command_group('deployment', resource_deployment_sdk, min_api='2018-05-01', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('list', 'list_deployments_at_subscription_scope', table_transformer=transform_deployments_list, deprecate_info=g.deprecate(redirect='deployment sub list', hide=True))
        g.custom_show_command('show', 'get_deployment_at_subscription_scope', deprecate_info=g.deprecate(redirect='deployment sub show', hide=True))
        g.custom_command('delete', 'delete_deployment_at_subscription_scope', supports_no_wait=True, deprecate_info=g.deprecate(redirect='deployment sub delete', hide=True))
        g.custom_command('validate', 'validate_arm_template_at_subscription_scope', validator=process_deployment_create_namespace,
                         table_transformer=deployment_validate_table_format, exception_handler=handle_template_based_exception,
                         deprecate_info=g.deprecate(redirect='deployment sub validate', hide=True))
        g.custom_command('create', 'deploy_arm_template_at_subscription_scope', supports_no_wait=True, validator=process_deployment_create_namespace,
                         exception_handler=handle_template_based_exception, deprecate_info=g.deprecate(redirect='deployment sub create', hide=True))
        g.custom_command('export', 'export_template_at_subscription_scope', deprecate_info=g.deprecate(redirect='deployment sub export', hide=True))
        g.custom_wait_command('wait', 'get_deployment_at_subscription_scope', deprecate_info=g.deprecate(redirect='deployment sub wait', hide=True))
        g.custom_command('cancel', 'cancel_deployment_at_subscription_scope', deprecate_info=g.deprecate(redirect='deployment sub cancel', hide=True))

    with self.command_group('deployment operation', resource_deployment_operation_sdk, min_api='2018-05-01', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('list', 'list_deployment_operations_at_subscription_scope',
                         deprecate_info=self.deprecate(redirect='deployment operation sub list', hide=True))
        g.custom_show_command('show', 'get_deployment_operations_at_subscription_scope', client_factory=cf_deployment_operations,
                              deprecate_info=self.deprecate(redirect='deployment operation sub show', hide=True))

    # az deployment sub
    with self.command_group('deployment sub', resource_deployment_sdk, min_api='2018-05-01', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('list', 'list_deployments_at_subscription_scope', table_transformer=transform_deployments_list)
        g.custom_show_command('show', 'get_deployment_at_subscription_scope', table_transformer=transform_deployment)
        g.custom_command('delete', 'delete_deployment_at_subscription_scope', supports_no_wait=True)
        g.custom_command('validate', 'validate_arm_template_at_subscription_scope', validator=process_deployment_create_namespace,
                         table_transformer=deployment_validate_table_format, exception_handler=handle_template_based_exception)
        g.custom_command('create', 'deploy_arm_template_at_subscription_scope', supports_no_wait=True, validator=process_deployment_create_namespace,
                         table_transformer=transform_deployment, exception_handler=handle_template_based_exception)
        g.custom_command('what-if', 'what_if_deploy_arm_template_at_subscription_scope', validator=process_deployment_create_namespace,
                         exception_handler=handle_template_based_exception, min_api='2019-07-01')
        g.custom_command('export', 'export_template_at_subscription_scope')
        g.custom_wait_command('wait', 'get_deployment_at_subscription_scope')
        g.custom_command('cancel', 'cancel_deployment_at_subscription_scope')

    with self.command_group('deployment operation sub', resource_deployment_operation_sdk, min_api='2018-05-01', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('list', 'list_deployment_operations_at_subscription_scope')
        g.custom_show_command('show', 'get_deployment_operations_at_subscription_scope', client_factory=cf_deployment_operations)

    with self.command_group('deployment-scripts', resource_deploymentscripts_sdk, resource_type=ResourceType.MGMT_RESOURCE_DEPLOYMENTSCRIPTS) as g:
        g.custom_command('list', 'list_deployment_scripts')
        g.custom_show_command('show', 'get_deployment_script')
        g.custom_command('show-log', 'get_deployment_script_logs')
        g.custom_command('delete', 'delete_deployment_script', confirmation=True)

    with self.command_group('ts', resource_templatespecs_sdk, resource_type=ResourceType.MGMT_RESOURCE_TEMPLATESPECS, min_api='2019-06-01-preview') as g:
        g.custom_command('create', 'create_template_spec', validator=process_ts_create_or_update_namespace)
        g.custom_command('update', 'update_template_spec', validator=process_ts_create_or_update_namespace, confirmation=True)
        g.custom_command('export', 'export_template_spec', validator=_validate_template_spec_out)
        g.custom_show_command('show', 'get_template_spec', validator=_validate_template_spec)
        g.custom_command('list', 'list_template_specs')
        g.custom_command('delete', 'delete_template_spec', validator=_validate_template_spec, confirmation=True)

    # az deployment group
    with self.command_group('deployment group', resource_deployment_sdk, resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('list', 'list_deployments_at_resource_group', table_transformer=transform_deployments_list)
        g.custom_show_command('show', 'get_deployment_at_resource_group', table_transformer=transform_deployment)
        g.custom_command('delete', 'delete_deployment_at_resource_group', supports_no_wait=True)
        g.custom_command('validate', 'validate_arm_template_at_resource_group', validator=process_deployment_create_namespace,
                         table_transformer=deployment_validate_table_format, exception_handler=handle_template_based_exception)
        g.custom_command('create', 'deploy_arm_template_at_resource_group', supports_no_wait=True, validator=process_deployment_create_namespace,
                         table_transformer=transform_deployment, exception_handler=handle_template_based_exception)
        g.custom_command('what-if', 'what_if_deploy_arm_template_at_resource_group', validator=process_deployment_create_namespace,
                         exception_handler=handle_template_based_exception, min_api='2019-07-01')
        g.custom_command('export', 'export_template_at_resource_group')
        g.custom_wait_command('wait', 'get_deployment_at_resource_group')
        g.custom_command('cancel', 'cancel_deployment_at_resource_group')

    with self.command_group('deployment operation group', resource_deployment_operation_sdk, resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('list', 'list_deployment_operations_at_resource_group')
        g.custom_show_command('show', 'get_deployment_operations_at_resource_group', client_factory=cf_deployment_operations)

    # az deployment mg
    with self.command_group('deployment mg', resource_deployment_sdk, min_api='2019-07-01', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('list', 'list_deployments_at_management_group', table_transformer=transform_deployments_list)
        g.custom_show_command('show', 'get_deployment_at_management_group', table_transformer=transform_deployment)
        g.custom_command('delete', 'delete_deployment_at_management_group', supports_no_wait=True)
        g.custom_command('validate', 'validate_arm_template_at_management_group', validator=process_deployment_create_namespace,
                         table_transformer=deployment_validate_table_format, exception_handler=handle_template_based_exception)
        g.custom_command('create', 'deploy_arm_template_at_management_group', supports_no_wait=True, validator=process_deployment_create_namespace,
                         table_transformer=transform_deployment, exception_handler=handle_template_based_exception)
        g.custom_command('what-if', 'what_if_deploy_arm_template_at_management_group', validator=process_deployment_create_namespace,
                         exception_handler=handle_template_based_exception, min_api='2019-10-01')
        g.custom_command('export', 'export_template_at_management_group')
        g.custom_wait_command('wait', 'get_deployment_at_management_group')
        g.custom_command('cancel', 'cancel_deployment_at_management_group')

    with self.command_group('deployment operation mg', resource_deployment_operation_sdk, min_api='2019-07-01', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('list', 'list_deployment_operations_at_management_group')
        g.custom_show_command('show', 'get_deployment_operations_at_management_group', client_factory=cf_deployment_operations)

    # az deployment tenant
    with self.command_group('deployment tenant', resource_deployment_sdk, min_api='2019-07-01', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('list', 'list_deployments_at_tenant_scope', table_transformer=transform_deployments_list)
        g.custom_show_command('show', 'get_deployment_at_tenant_scope', table_transformer=transform_deployment)
        g.custom_command('delete', 'delete_deployment_at_tenant_scope', supports_no_wait=True)
        g.custom_command('validate', 'validate_arm_template_at_tenant_scope', validator=process_deployment_create_namespace,
                         table_transformer=deployment_validate_table_format, exception_handler=handle_template_based_exception)
        g.custom_command('create', 'deploy_arm_template_at_tenant_scope', supports_no_wait=True, validator=process_deployment_create_namespace,
                         table_transformer=transform_deployment, exception_handler=handle_template_based_exception)
        g.custom_command('what-if', 'what_if_deploy_arm_template_at_tenant_scope', validator=process_deployment_create_namespace,
                         exception_handler=handle_template_based_exception, min_api='2019-10-01')
        g.custom_command('export', 'export_template_at_tenant_scope')
        g.custom_wait_command('wait', 'get_deployment_at_tenant_scope')
        g.custom_command('cancel', 'cancel_deployment_at_tenant_scope')

    with self.command_group('deployment operation tenant', resource_deployment_operation_sdk, min_api='2019-07-01', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('list', 'list_deployment_operations_at_tenant_scope')
        g.custom_show_command('show', 'get_deployment_operations_at_tenant_scope', client_factory=cf_deployment_operations)

    with self.command_group('policy assignment', resource_type=ResourceType.MGMT_RESOURCE_POLICY) as g:
        g.custom_command('create', 'create_policy_assignment')
        g.custom_command('delete', 'delete_policy_assignment')
        g.custom_command('list', 'list_policy_assignment')
        g.custom_show_command('show', 'show_policy_assignment')
        g.custom_command('update', 'update_policy_assignment')

    with self.command_group('policy assignment identity', resource_type=ResourceType.MGMT_RESOURCE_POLICY, min_api='2018-05-01') as g:
        g.custom_command('assign', 'set_identity')
        g.custom_show_command('show', 'show_identity')
        g.custom_command('remove', 'remove_identity')

    with self.command_group('policy assignment non-compliance-message', resource_type=ResourceType.MGMT_RESOURCE_POLICY, min_api='2020-09-01') as g:
        g.custom_command('create', 'create_policy_non_compliance_message')
        g.custom_command('list', 'list_policy_non_compliance_message')
        g.custom_command('delete', 'delete_policy_non_compliance_message')

    with self.command_group('policy definition', resource_policy_definitions_sdk, resource_type=ResourceType.MGMT_RESOURCE_POLICY) as g:
        g.custom_command('create', 'create_policy_definition')
        g.custom_command('delete', 'delete_policy_definition')
        g.custom_command('list', 'list_policy_definition')
        g.custom_show_command('show', 'get_policy_definition')
        g.custom_command('update', 'update_policy_definition')

    with self.command_group('policy set-definition', resource_policy_set_definitions_sdk, resource_type=ResourceType.MGMT_RESOURCE_POLICY, min_api='2017-06-01-preview') as g:
        g.custom_command('create', 'create_policy_setdefinition')
        g.custom_command('delete', 'delete_policy_setdefinition')
        g.custom_command('list', 'list_policy_setdefinition')
        g.custom_show_command('show', 'get_policy_setdefinition')
        g.custom_command('update', 'update_policy_setdefinition')

    with self.command_group('policy exemption', resource_policy_exemptions_sdk, is_preview=True, resource_type=ResourceType.MGMT_RESOURCE_POLICY, min_api='2020-09-01') as g:
        g.custom_command('create', 'create_policy_exemption')
        g.custom_command('delete', 'delete_policy_exemption')
        g.custom_command('list', 'list_policy_exemption')
        g.custom_show_command('show', 'get_policy_exemption')
        g.custom_command('update', 'update_policy_exemption')

    with self.command_group('lock', resource_type=ResourceType.MGMT_RESOURCE_LOCKS) as g:
        g.custom_command('create', 'create_lock')
        g.custom_command('delete', 'delete_lock')
        g.custom_command('list', 'list_locks')
        g.custom_show_command('show', 'get_lock')
        g.custom_command('update', 'update_lock')

    with self.command_group('resource link', resource_link_sdk, resource_type=ResourceType.MGMT_RESOURCE_LINKS) as g:
        g.custom_command('create', 'create_resource_link')
        g.command('delete', 'delete')
        g.show_command('show', 'get')
        g.custom_command('list', 'list_resource_links')
        g.custom_command('update', 'update_resource_link')

    with self.command_group('managedapp', resource_managedapp_sdk, min_api='2017-05-10', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('create', 'create_application')
        g.command('delete', 'begin_delete')
        g.custom_show_command('show', 'show_application')
        g.custom_command('list', 'list_applications')

    with self.command_group('managedapp definition', resource_managedapp_def_sdk, min_api='2017-05-10', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES) as g:
        g.custom_command('create', 'create_or_update_applicationdefinition')
        g.custom_command('update', 'create_or_update_applicationdefinition')
        g.command('delete', 'begin_delete')
        g.custom_show_command('show', 'show_applicationdefinition')
        g.command('list', 'list_by_resource_group', exception_handler=empty_on_404)

    with self.command_group('account management-group', resource_managementgroups_sdk, client_factory=cf_management_groups) as g:
        g.custom_command('list', 'cli_managementgroups_group_list')
        g.custom_show_command('show', 'cli_managementgroups_group_show')
        g.custom_command('create', 'cli_managementgroups_group_create')
        g.custom_command('delete', 'cli_managementgroups_group_delete')
        g.generic_update_command(
            'update',
            getter_name='cli_managementgroups_group_update_get',
            getter_type=resource_managementgroups_update_type,
            setter_name='cli_managementgroups_group_update_set',
            setter_type=resource_managementgroups_update_type,
            custom_func_name='cli_managementgroups_group_update_custom_func',
            custom_func_type=resource_managementgroups_update_type,
            exception_handler=managementgroups_exception_handler)

    with self.command_group('account management-group subscription', resource_managementgroups_subscriptions_sdk, client_factory=cf_management_group_subscriptions) as g:
        g.custom_command('add', 'cli_managementgroups_subscription_add')
        g.custom_command('remove', 'cli_managementgroups_subscription_remove')

    with self.command_group('bicep') as g:
        g.custom_command('install', 'install_bicep_cli')
        g.custom_command('uninstall', 'uninstall_bicep_cli')
        g.custom_command('upgrade', 'upgrade_bicep_cli')
        g.custom_command('build', 'build_bicep_file')
        g.custom_command('decompile', 'decompile_bicep_file')
        g.custom_command('version', 'show_bicep_cli_version')
        g.custom_command('list-versions', 'list_bicep_cli_versions')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from azure.cli.command_modules.monitor._client_factory import _log_analytics_client_factory
from azure.cli.core.commands.transform import _parse_id
from azure.mgmt.loganalytics.models import WorkspaceSkuNameEnum, Workspace, WorkspaceSku, WorkspaceCapping, Table
from azure.cli.core.util import sdk_no_wait
from knack.util import CLIError


def create_log_analytics_workspace(client, resource_group_name, workspace_name, location=None, tags=None,
                                   sku=WorkspaceSkuNameEnum.per_gb2018.value, capacity_reservation_level=None,
                                   retention_time=None, daily_quota_gb=None,
                                   public_network_access_for_query=None, public_network_access_for_ingestion=None,
                                   no_wait=False):
    if sku.lower() == WorkspaceSkuNameEnum.capacity_reservation.value.lower() and capacity_reservation_level is None:
        raise CLIError('--capacity-reservation-level must be set when sku is CapacityReservation.')
    if sku.lower() != WorkspaceSkuNameEnum.capacity_reservation.value.lower() and capacity_reservation_level:
        raise CLIError('--capacity-reservation-level can be set only when sku is CapacityReservation.')

    workspace_client = client
    sku = WorkspaceSku(name=sku, capacity_reservation_level=capacity_reservation_level)
    workspace_capping = WorkspaceCapping(daily_quota_gb=daily_quota_gb)
    workspace_instance = Workspace(location=location,
                                   tags=tags,
                                   sku=sku,
                                   workspace_capping=workspace_capping,
                                   retention_in_days=retention_time,
                                   public_network_access_for_query=public_network_access_for_query,
                                   public_network_access_for_ingestion=public_network_access_for_ingestion)
    return sdk_no_wait(no_wait, workspace_client.begin_create_or_update, resource_group_name,
                       workspace_name, workspace_instance)


def update_log_analytics_workspace(instance, tags=None, capacity_reservation_level=None,
                                   retention_time=None, daily_quota_gb=None,
                                   public_network_access_for_query=None, public_network_access_for_ingestion=None):
    if tags is not None:
        instance.tags = tags
    if capacity_reservation_level is not None:
        if instance.sku.name.lower() != WorkspaceSkuNameEnum.capacity_reservation.value.lower():
            raise CLIError('--capacity-reservation-level can be set only when sku is CapacityReservation.')
        instance.sku.capacity_reservation_level = capacity_reservation_level
    if retention_time is not None:
        instance.retention_in_days = retention_time
    if daily_quota_gb is not None:
        instance.workspace_capping.daily_quota_gb = daily_quota_gb
    if public_network_access_for_query is not None:
        instance.public_network_access_for_query = public_network_access_for_query
    if public_network_access_for_ingestion is not None:
        instance.public_network_access_for_ingestion = public_network_access_for_ingestion
    return instance


def list_log_analytics_workspace(client, resource_group_name=None):
    workspace_client = client
    if resource_group_name:
        return workspace_client.list_by_resource_group(resource_group_name)
    return workspace_client.list()


def list_deleted_log_analytics_workspaces(client, resource_group_name=None):
    if resource_group_name:
        return client.list_by_resource_group(resource_group_name)
    return client.list()


def recover_log_analytics_workspace(cmd, workspace_name, resource_group_name=None, no_wait=False):
    deleted_workspaces_client = _log_analytics_client_factory(cmd.cli_ctx).deleted_workspaces
    workspace_client = _log_analytics_client_factory(cmd.cli_ctx).workspaces
    deleted_workspaces = list(list_deleted_log_analytics_workspaces(deleted_workspaces_client, resource_group_name))
    for deleted_workspace in deleted_workspaces:
        if deleted_workspace.name.lower() == workspace_name.lower():
            workspace_instance = Workspace(location=deleted_workspace.location)

            return sdk_no_wait(no_wait, workspace_client.begin_create_or_update,
                               _parse_id(deleted_workspace.id)['resource-group'],
                               workspace_name, workspace_instance)

    raise CLIError('{} is not a deleted workspace and you can only recover a deleted workspace within 14 days.'
                   .format(workspace_name))


def _format_tags(tags):
    from azure.mgmt.loganalytics.models import Tag
    if tags:
        tags = [Tag(name=key, value=value) for key, value in tags.items()]
    return tags


def create_log_analytics_workspace_saved_search(client, workspace_name, resource_group_name, saved_search_id,
                                                category, display_name, saved_query,
                                                function_alias=None, function_parameters=None,
                                                tags=None):
    from azure.mgmt.loganalytics.models import SavedSearch
    saved_search = SavedSearch(category=category,
                               display_name=display_name,
                               query=saved_query,
                               function_alias=function_alias,
                               function_parameters=function_parameters,
                               tags=_format_tags(tags))
    return client.create_or_update(resource_group_name=resource_group_name,
                                   workspace_name=workspace_name,
                                   saved_search_id=saved_search_id,
                                   parameters=saved_search)


def update_log_analytics_workspace_saved_search(cmd, instance, category=None, display_name=None, saved_query=None,
                                                function_alias=None, function_parameters=None,
                                                tags=None):
    with cmd.update_context(instance) as c:
        c.set_param('category', category)
        c.set_param('display_name', display_name)
        c.set_param('query', saved_query)
        c.set_param('function_alias', function_alias)
        c.set_param('function_parameters', function_parameters)
        c.set_param('tags', _format_tags(tags))
    return instance


def create_log_analytics_workspace_data_exports(client, workspace_name, resource_group_name, data_export_name,
                                                destination, data_export_type, table_names,
                                                event_hub_name=None, enable=None):
    from azure.mgmt.loganalytics.models import DataExport
    data_export = DataExport(resource_id=destination,
                             data_export_type=data_export_type,
                             table_names=table_names,
                             event_hub_name=event_hub_name,
                             enable=enable)
    return client.create_or_update(resource_group_name, workspace_name, data_export_name, data_export)


def update_log_analytics_workspace_data_exports(cmd, instance, table_names, destination=None, data_export_type=None,
                                                event_hub_name=None, enable=None):
    with cmd.update_context(instance) as c:
        c.set_param('resource_id', destination)
        c.set_param('data_export_type', data_export_type)
        c.set_param('table_names', table_names)
        c.set_param('event_hub_name', event_hub_name)
        c.set_param('enable', enable)
    return instance


def update_log_analytics_workspace_table(client, resource_group_name, workspace_name, table_name, retention_in_days):
    table = Table(retention_in_days=retention_in_days)
    return client.update(resource_group_name=resource_group_name,
                         workspace_name=workspace_name,
                         table_name=table_name,
                         parameters=table)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=unused-argument, line-too-long
from importlib import import_module
from msrestazure.azure_exceptions import CloudError
from msrestazure.tools import resource_id, is_valid_resource_id, parse_resource_id  # pylint: disable=import-error
from knack.log import get_logger
from azure.core.exceptions import ResourceNotFoundError
from azure.cli.core.azclierror import RequiredArgumentMissingError, ArgumentUsageError, InvalidArgumentValueError
from azure.cli.core.commands.client_factory import get_subscription_id
from azure.cli.core.util import CLIError, sdk_no_wait, user_confirmation
from azure.cli.core.local_context import ALL
from azure.mgmt.rdbms import mysql_flexibleservers
from ._client_factory import get_mysql_flexible_management_client, cf_mysql_flexible_firewall_rules, \
    cf_mysql_flexible_db, cf_mysql_check_resource_availability, cf_mysql_flexible_private_dns_zone_suffix_operations
from ._flexible_server_util import resolve_poller, generate_missing_parameters, get_mysql_list_skus_info, \
    generate_password, parse_maintenance_window
from .flexible_server_custom_common import create_firewall_rule
from .flexible_server_virtual_network import prepare_private_network, prepare_private_dns_zone, prepare_public_network
from .validators import mysql_arguments_validator

logger = get_logger(__name__)
DEFAULT_DB_NAME = 'flexibleserverdb'
DELEGATION_SERVICE_NAME = "Microsoft.DBforMySQL/flexibleServers"
MINIMUM_IOPS = 300
RESOURCE_PROVIDER = 'Microsoft.DBforMySQL'


# region create without args
# pylint: disable=too-many-locals, too-many-statements, raise-missing-from
def flexible_server_create(cmd, client,
                           resource_group_name=None, server_name=None,
                           location=None, backup_retention=None,
                           sku_name=None, tier=None,
                           storage_gb=None, administrator_login=None,
                           administrator_login_password=None, version=None,
                           tags=None, database_name=None,
                           subnet=None, subnet_address_prefix=None, vnet=None, vnet_address_prefix=None,
                           private_dns_zone_arguments=None, public_access=None,
                           high_availability=None, zone=None, standby_availability_zone=None,
                           iops=None, auto_grow=None, yes=False):
    # Generate missing parameters
    location, resource_group_name, server_name = generate_missing_parameters(cmd, location, resource_group_name,
                                                                             server_name, 'mysql')
    db_context = DbContext(
        cmd=cmd, cf_firewall=cf_mysql_flexible_firewall_rules, cf_db=cf_mysql_flexible_db,
        cf_availability=cf_mysql_check_resource_availability, cf_private_dns_zone_suffix=cf_mysql_flexible_private_dns_zone_suffix_operations, logging_name='MySQL', command_group='mysql', server_client=client,
        location=location)

    server_name = server_name.lower()

    mysql_arguments_validator(db_context,
                              server_name=server_name,
                              location=location,
                              tier=tier,
                              sku_name=sku_name,
                              storage_gb=storage_gb,
                              backup_retention=backup_retention,
                              high_availability=high_availability,
                              standby_availability_zone=standby_availability_zone,
                              zone=zone,
                              subnet=subnet,
                              public_access=public_access,
                              auto_grow=auto_grow,
                              version=version)
    _, _, iops_info = get_mysql_list_skus_info(db_context.cmd, location)

    server_result = firewall_id = subnet_id = None

    network = mysql_flexibleservers.models.Network()
    if subnet is not None or vnet is not None:
        subnet_id = prepare_private_network(cmd,
                                            resource_group_name,
                                            server_name,
                                            vnet=vnet,
                                            subnet=subnet,
                                            location=location,
                                            delegation_service_name=DELEGATION_SERVICE_NAME,
                                            vnet_address_pref=vnet_address_prefix,
                                            subnet_address_pref=subnet_address_prefix,
                                            yes=yes)
        private_dns_zone_id = prepare_private_dns_zone(db_context,
                                                       'MySQL',
                                                       resource_group_name,
                                                       server_name,
                                                       private_dns_zone=private_dns_zone_arguments,
                                                       subnet_id=subnet_id,
                                                       location=location,
                                                       yes=yes)
        network.delegated_subnet_resource_id = subnet_id
        network.private_dns_zone_resource_id = private_dns_zone_id
    else:
        network.public_network_access = ' Enabled'
        start_ip, end_ip = prepare_public_network(public_access, yes=yes)
        if start_ip != -1:
            public_access = 'Enabled'

    # determine IOPS
    iops = _determine_iops(storage_gb=storage_gb,
                           iops_info=iops_info,
                           iops_input=iops,
                           tier=tier,
                           sku_name=sku_name)

    storage = mysql_flexibleservers.models.Storage(storage_size_gb=storage_gb,
                                                   iops=iops,
                                                   auto_grow=auto_grow)

    backup = mysql_flexibleservers.models.Backup(backup_retention_days=backup_retention)

    sku = mysql_flexibleservers.models.Sku(name=sku_name, tier=tier)

    if high_availability.lower() == "enabled":
        high_availability = "ZoneRedundant"
    high_availability = mysql_flexibleservers.models.HighAvailability(mode=high_availability,
                                                                      standby_availability_zone=standby_availability_zone)

    administrator_login_password = generate_password(administrator_login_password)

    # Create mysql server
    # Note : passing public_access has no effect as the accepted values are 'Enabled' and 'Disabled'. So the value ends up being ignored.
    server_result = _create_server(db_context, cmd, resource_group_name, server_name,
                                   tags=tags,
                                   location=location,
                                   sku=sku,
                                   administrator_login=administrator_login,
                                   administrator_login_password=administrator_login_password,
                                   storage=storage,
                                   backup=backup,
                                   network=network,
                                   version=version,
                                   high_availability=high_availability,
                                   availability_zone=zone)

    # Adding firewall rule
    if public_access is not None and str(public_access).lower() != 'none':
        firewall_id = create_firewall_rule(db_context, cmd, resource_group_name, server_name, start_ip, end_ip)

    # Create mysql database if it does not exist
    if database_name is None:
        database_name = DEFAULT_DB_NAME
    _create_database(db_context, cmd, resource_group_name, server_name, database_name)

    user = server_result.administrator_login
    server_id = server_result.id
    loc = server_result.location
    version = server_result.version
    sku = server_result.sku.name
    host = server_result.fully_qualified_domain_name

    logger.warning('Make a note of your password. If you forget, you would have to reset your password with'
                   '\'az mysql flexible-server update -n %s -g %s -p <new-password>\'.',
                   server_name, resource_group_name)

    _update_local_contexts(cmd, server_name, resource_group_name, location, user)

    return _form_response(user, sku, loc, server_id, host, version,
                          administrator_login_password if administrator_login_password is not None else '*****',
                          _create_mysql_connection_string(host, database_name, user, administrator_login_password),
                          database_name, firewall_id, subnet_id)


def flexible_server_restore(cmd, client,
                            resource_group_name, server_name,
                            source_server, restore_point_in_time=None, zone=None, no_wait=False,
                            subnet=None, subnet_address_prefix=None, vnet=None, vnet_address_prefix=None,
                            private_dns_zone_arguments=None, yes=False):
    provider = 'Microsoft.DBforMySQL'
    server_name = server_name.lower()

    if not is_valid_resource_id(source_server):
        if len(source_server.split('/')) == 1:
            source_server_id = resource_id(
                subscription=get_subscription_id(cmd.cli_ctx),
                resource_group=resource_group_name,
                namespace=provider,
                type='flexibleServers',
                name=source_server)
        else:
            raise ValueError('The provided source-server {} is invalid.'.format(source_server))
    else:
        source_server_id = source_server

    try:
        id_parts = parse_resource_id(source_server_id)
        source_server_object = client.get(id_parts['resource_group'], id_parts['name'])

        if not zone:
            zone = source_server_object.availability_zone

        parameters = mysql_flexibleservers.models.Server(
            location=source_server_object.location,
            restore_point_in_time=restore_point_in_time,
            source_server_resource_id=source_server_id,  # this should be the source server name, not id
            create_mode="PointInTimeRestore",
            availability_zone=zone
        )

        if source_server_object.network.public_network_access == 'Disabled':
            db_context = DbContext(
                cmd=cmd, cf_firewall=cf_mysql_flexible_firewall_rules, cf_db=cf_mysql_flexible_db,
                cf_availability=cf_mysql_check_resource_availability, cf_private_dns_zone_suffix=cf_mysql_flexible_private_dns_zone_suffix_operations, logging_name='MySQL', command_group='mysql', server_client=client,
                location=source_server_object.location)

            if subnet is not None or vnet is not None:
                network = mysql_flexibleservers.models.Network()
                subnet_id = prepare_private_network(cmd,
                                                    resource_group_name,
                                                    server_name,
                                                    vnet=vnet,
                                                    subnet=subnet,
                                                    location=source_server_object.location,
                                                    delegation_service_name=DELEGATION_SERVICE_NAME,
                                                    vnet_address_pref=vnet_address_prefix,
                                                    subnet_address_pref=subnet_address_prefix,
                                                    yes=yes)
                private_dns_zone_id = prepare_private_dns_zone(db_context,
                                                               'MySQL',
                                                               resource_group_name,
                                                               server_name,
                                                               private_dns_zone=private_dns_zone_arguments,
                                                               subnet_id=subnet_id,
                                                               location=source_server_object.location,
                                                               yes=yes)
                network.delegated_subnet_resource_id = subnet_id
                network.private_dns_zone_resource_id = private_dns_zone_id
                parameters.network = network

    except Exception as e:
        raise ResourceNotFoundError(e)

    return sdk_no_wait(no_wait, client.begin_create, resource_group_name, server_name, parameters)


# pylint: disable=too-many-branches
def flexible_server_update_custom_func(cmd, client, instance,
                                       sku_name=None,
                                       tier=None,
                                       storage_gb=None,
                                       auto_grow=None,
                                       iops=None,
                                       backup_retention=None,
                                       administrator_login_password=None,
                                       high_availability=None,
                                       standby_availability_zone=None,
                                       maintenance_window=None,
                                       tags=None,
                                       replication_role=None):
    # validator
    location = ''.join(instance.location.lower().split())
    db_context = DbContext(
        cmd=cmd, cf_firewall=cf_mysql_flexible_firewall_rules, cf_db=cf_mysql_flexible_db,
        cf_availability=cf_mysql_check_resource_availability, logging_name='MySQL', command_group='mysql', server_client=client,
        location=instance.location)

    mysql_arguments_validator(db_context,
                              location=location,
                              tier=tier,
                              sku_name=sku_name,
                              storage_gb=storage_gb,
                              backup_retention=backup_retention,
                              high_availability=high_availability,
                              zone=instance.availability_zone,
                              standby_availability_zone=standby_availability_zone,
                              auto_grow=auto_grow,
                              replication_role=replication_role,
                              instance=instance)
    _, _, iops_info = get_mysql_list_skus_info(db_context.cmd, location)

    server_module_path = instance.__module__
    module = import_module(server_module_path)  # replacement not needed for update in flex servers
    ServerForUpdate = getattr(module, 'ServerForUpdate')

    if sku_name:
        instance.sku.name = sku_name

    if tier:
        instance.sku.tier = tier

    if backup_retention:
        instance.backup.backup_retention_days = backup_retention

    if maintenance_window:
        # if disabled is pass in reset to default values
        if maintenance_window.lower() == "disabled":
            day_of_week = start_hour = start_minute = 0
            custom_window = "Disabled"
        else:
            day_of_week, start_hour, start_minute = parse_maintenance_window(maintenance_window)
            custom_window = "Enabled"

        # set values - if maintenance_window when is None when created then create a new object
        if instance.maintenance_window is None:
            instance.maintenance_window = mysql_flexibleservers.models.MaintenanceWindow(
                day_of_week=day_of_week,
                start_hour=start_hour,
                start_minute=start_minute,
                custom_window=custom_window
            )
        else:
            instance.maintenance_window.day_of_week = day_of_week
            instance.maintenance_window.start_hour = start_hour
            instance.maintenance_window.start_minute = start_minute
            instance.maintenance_window.custom_window = custom_window

    if high_availability:
        if high_availability.lower() == "enabled":
            high_availability = "ZoneRedundant"
            instance.high_availability.mode = high_availability
            if standby_availability_zone:
                instance.high_availability.standby_availability_zone = standby_availability_zone
        else:
            instance.high_availability = mysql_flexibleservers.models.HighAvailability(mode=high_availability)

    if storage_gb:
        instance.storage.storage_size_gb = storage_gb

    if not iops:
        iops = instance.storage.iops
    instance.storage.iops = _determine_iops(storage_gb=instance.storage.storage_size_gb,
                                            iops_info=iops_info,
                                            iops_input=iops,
                                            tier=instance.sku.tier,
                                            sku_name=instance.sku.name)

    if auto_grow:
        instance.storage.storage_autogrow = auto_grow

    params = ServerForUpdate(sku=instance.sku,
                             storage=instance.storage,
                             backup=instance.backup,
                             administrator_login_password=administrator_login_password,
                             high_availability=instance.high_availability,
                             maintenance_window=instance.maintenance_window,
                             tags=tags)

    return params


def server_delete_func(cmd, client, resource_group_name=None, server_name=None, yes=None):
    result = None  # default return value

    if not yes:
        user_confirmation(
            "Are you sure you want to delete the server '{0}' in resource group '{1}'".format(server_name,
                                                                                              resource_group_name), yes=yes)
    try:
        result = client.begin_delete(resource_group_name, server_name)
        if cmd.cli_ctx.local_context.is_on:
            local_context_file = cmd.cli_ctx.local_context._get_local_context_file()  # pylint: disable=protected-access
            local_context_file.remove_option('mysql flexible-server', 'server_name')
            local_context_file.remove_option('mysql flexible-server', 'administrator_login')
            local_context_file.remove_option('mysql flexible-server', 'database_name')

    except Exception as ex:  # pylint: disable=broad-except
        logger.error(ex)
        raise CLIError(ex)
    return result


def flexible_server_restart(cmd, client, resource_group_name, server_name, fail_over=None):
    instance = client.get(resource_group_name, server_name)
    if fail_over is not None and instance.high_availability.mode not in ["ZoneRedundant", "Enabled"]:
        raise ArgumentUsageError("Failing over can only be triggered for zone redundant servers.")

    if fail_over is not None:
        if fail_over != 'Forced':
            raise InvalidArgumentValueError("Allowed failover parameters are 'Forced'.")
        parameters = mysql_flexibleservers.models.ServerRestartParameter(restart_with_failover='Enabled')
    else:
        parameters = mysql_flexibleservers.models.ServerRestartParameter(restart_with_failover='Disabled')

    return resolve_poller(
        client.begin_restart(resource_group_name, server_name, parameters), cmd.cli_ctx, 'MySQL Server Restart')


# Parameter update command
def flexible_parameter_update(client, server_name, configuration_name, resource_group_name, source=None, value=None):
    if source is None and value is None:
        # update the command with system default
        try:
            parameter = client.get(resource_group_name, server_name, configuration_name)
            value = parameter.default_value  # reset value to default
            source = "system-default"
        except CloudError as e:
            raise CLIError('Unable to get default parameter value: {}.'.format(str(e)))
    elif source is None:
        source = "user-override"

    parameters = mysql_flexibleservers.models.Configuration(
        name=configuration_name,
        value=value,
        source=source
    )

    return client.begin_update(resource_group_name, server_name, configuration_name, parameters)


# Replica commands
# Custom functions for server replica, will add MySQL part after backend ready in future
def flexible_replica_create(cmd, client, resource_group_name, source_server, replica_name, no_wait=False, location=None, sku_name=None, tier=None, **kwargs):
    provider = 'Microsoft.DBforMySQL'
    replica_name = replica_name.lower()

    # set source server id
    if not is_valid_resource_id(source_server):
        if len(source_server.split('/')) == 1:
            source_server_id = resource_id(subscription=get_subscription_id(cmd.cli_ctx),
                                           resource_group=resource_group_name,
                                           namespace=provider,
                                           type='flexibleServers',
                                           name=source_server)
        else:
            raise CLIError('The provided source-server {} is invalid.'.format(source_server))
    else:
        source_server_id = source_server

    source_server_id_parts = parse_resource_id(source_server_id)
    try:
        source_server_object = client.get(source_server_id_parts['resource_group'], source_server_id_parts['name'])
    except Exception as e:
        raise ResourceNotFoundError(e)

    location = source_server_object.location
    sku_name = source_server_object.sku.name
    tier = source_server_object.sku.tier

    parameters = mysql_flexibleservers.models.Server(
        sku=mysql_flexibleservers.models.Sku(name=sku_name, tier=tier),
        source_server_resource_id=source_server_id,
        location=location,
        create_mode="Replica")
    return sdk_no_wait(no_wait, client.begin_create, resource_group_name, replica_name, parameters)


def flexible_replica_stop(client, resource_group_name, server_name):
    try:
        server_object = client.get(resource_group_name, server_name)
    except Exception as e:
        raise ResourceNotFoundError(e)

    if server_object.replication_role is not None and server_object.replication_role.lower() != "replica":
        raise CLIError('Server {} is not a replica server.'.format(server_name))

    server_module_path = server_object.__module__
    module = import_module(server_module_path)  # replacement not needed for update in flex servers
    ServerForUpdate = getattr(module, 'ServerForUpdate')

    params = ServerForUpdate(replication_role='None')

    return client.begin_update(resource_group_name, server_name, params)


def flexible_server_mysql_get(cmd, resource_group_name, server_name):
    client = get_mysql_flexible_management_client(cmd.cli_ctx)
    return client.servers.get(resource_group_name, server_name)


def flexible_list_skus(cmd, client, location):
    result = client.list(location)
    logger.warning('For prices please refer to https://aka.ms/mysql-pricing')
    return result


def _create_server(db_context, cmd, resource_group_name, server_name, tags, location, sku, administrator_login, administrator_login_password,
                   storage, backup, network, version, high_availability, availability_zone):
    logging_name, server_client = db_context.logging_name, db_context.server_client
    logger.warning('Creating %s Server \'%s\' in group \'%s\'...', logging_name, server_name, resource_group_name)

    logger.warning('Your server \'%s\' is using sku \'%s\' (Paid Tier). '
                   'Please refer to https://aka.ms/mysql-pricing for pricing details', server_name, sku.name)
    # Note : passing public-network-access has no effect as the accepted values are 'Enabled' and 'Disabled'.
    # So when you pass an IP here(from the CLI args of public_access), it ends up being ignored.
    parameters = mysql_flexibleservers.models.Server(
        tags=tags,
        location=location,
        sku=sku,
        administrator_login=administrator_login,
        administrator_login_password=administrator_login_password,
        storage=storage,
        backup=backup,
        network=network,
        version=version,
        high_availability=high_availability,
        availability_zone=availability_zone,
        create_mode="Create")

    # if assign_identity:
    #     parameters.identity = mysql_flexibleservers.models.Identity()

    return resolve_poller(
        server_client.begin_create(resource_group_name, server_name, parameters), cmd.cli_ctx,
        '{} Server Create'.format(logging_name))


def flexible_server_connection_string(
        server_name='{server}', database_name='{database}', administrator_login='{login}',
        administrator_login_password='{password}'):
    host = '{}.mysql.database.azure.com'.format(server_name)
    if database_name is None:
        database_name = 'mysql'
    return {
        'connectionStrings': _create_mysql_connection_strings(host, administrator_login, administrator_login_password,
                                                              database_name)
    }


def _create_mysql_connection_strings(host, user, password, database):
    result = {
        'mysql_cmd': "mysql {database} --host {host} --user {user} --password={password}",
        'ado.net': "Server={host}; Port=3306; Database={database}; Uid={user}; Pwd={password};",
        'jdbc': "jdbc:mysql://{host}:3306/{database}?user={user}&password={password}",
        'jdbc Spring': "spring.datasource.url=jdbc:mysql://{host}:3306/{database}  "
                       "spring.datasource.username={user}  "
                       "spring.datasource.password={password}",
        'node.js': "var conn = mysql.createConnection({{host: '{host}', user: '{user}', "
                   "password: {password}, database: {database}, port: 3306}});",
        'php': "host={host} port=3306 dbname={database} user={user} password={password}",
        'python': "cnx = mysql.connector.connect(user='{user}', password='{password}', host='{host}', "
                  "port=3306, database='{database}')",
        'ruby': "client = Mysql2::Client.new(username: '{user}', password: '{password}', "
                "database: '{database}', host: '{host}', port: 3306)",
    }

    connection_kwargs = {
        'host': host,
        'user': user,
        'password': password if password is not None else '{password}',
        'database': database
    }

    for k, v in result.items():
        result[k] = v.format(**connection_kwargs)
    return result


def _form_response(username, sku, location, server_id, host, version, password, connection_string, database_name,
                   firewall_id=None, subnet_id=None):
    output = {
        'host': host,
        'username': username,
        'password': password,
        'skuname': sku,
        'location': location,
        'id': server_id,
        'version': version,
        'databaseName': database_name,
        'connectionString': connection_string
    }
    if firewall_id is not None:
        output['firewallName'] = firewall_id
    if subnet_id is not None:
        output['subnetId'] = subnet_id
    return output


def _update_local_contexts(cmd, server_name, resource_group_name, location, user):
    if cmd.cli_ctx.local_context.is_on:
        cmd.cli_ctx.local_context.set(['mysql flexible-server'], 'server_name',
                                      server_name)  # Setting the server name in the local context
        cmd.cli_ctx.local_context.set([ALL], 'location',
                                      location)  # Setting the location in the local context
        cmd.cli_ctx.local_context.set([ALL], 'resource_group_name', resource_group_name)
        cmd.cli_ctx.local_context.set(['mysql flexible-server'], 'administrator_login',
                                      user)  # Setting the server name in the local context


def _create_database(db_context, cmd, resource_group_name, server_name, database_name):
    # check for existing database, create if not
    cf_db, logging_name = db_context.cf_db, db_context.logging_name
    database_client = cf_db(cmd.cli_ctx, None)

    logger.warning('Creating %s database \'%s\'...', logging_name, database_name)
    parameters = {
        'name': database_name,
        'charset': 'utf8',
        'collation': 'utf8_general_ci'
    }
    resolve_poller(
        database_client.begin_create_or_update(resource_group_name, server_name, database_name, parameters), cmd.cli_ctx,
        '{} Database Create/Update'.format(logging_name))


def database_create_func(client, resource_group_name=None, server_name=None, database_name=None, charset=None, collation=None):

    if charset is None and collation is None:
        charset = 'utf8'
        collation = 'utf8_general_ci'
        logger.warning("Creating database with utf8 charset and utf8_general_ci collation")
    elif (not charset and collation) or (charset and not collation):
        raise RequiredArgumentMissingError("charset and collation have to be input together.")

    parameters = {
        'name': database_name,
        'charset': charset,
        'collation': collation
    }

    return client.begin_create_or_update(
        resource_group_name,
        server_name,
        database_name,
        parameters)


def _create_mysql_connection_string(host, database_name, user_name, password):
    connection_kwargs = {
        'host': host,
        'dbname': database_name,
        'username': user_name,
        'password': password if password is not None else '{password}'
    }
    return 'mysql {dbname} --host {host} --user {username} --password={password}'.format(**connection_kwargs)


def _determine_iops(storage_gb, iops_info, iops_input, tier, sku_name):
    max_supported_iops = iops_info[tier][sku_name]
    free_iops = get_free_iops(storage_in_mb=storage_gb * 1024,
                              iops_info=iops_info,
                              tier=tier,
                              sku_name=sku_name)

    iops = free_iops
    if iops_input and iops_input > free_iops:
        iops = min(iops_input, max_supported_iops)

    logger.warning("IOPS is %d which is either your input or free(maximum) IOPS supported for your storage size and SKU.", iops)
    return iops


def get_free_iops(storage_in_mb, iops_info, tier, sku_name):
    free_iops = MINIMUM_IOPS + (storage_in_mb // 1024) * 3
    max_supported_iops = iops_info[tier][sku_name]  # free iops cannot exceed maximum supported iops for the sku

    return min(free_iops, max_supported_iops)


# pylint: disable=too-many-instance-attributes, too-few-public-methods, useless-object-inheritance
class DbContext(object):
    def __init__(self, cmd=None, azure_sdk=None, logging_name=None, cf_firewall=None, cf_db=None,
                 cf_availability=None, cf_private_dns_zone_suffix=None, command_group=None, server_client=None, location=None):
        self.cmd = cmd
        self.azure_sdk = azure_sdk
        self.cf_firewall = cf_firewall
        self.cf_db = cf_db
        self.cf_availability = cf_availability
        self.cf_private_dns_zone_suffix = cf_private_dns_zone_suffix
        self.logging_name = logging_name
        self.command_group = command_group
        self.server_client = server_client
        self.location = location



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.log import get_logger
from knack.util import CLIError
from azure.cli.core.commands import LongRunningOperation

from ._constants import ACR_TASK_YAML_DEFAULT_NAME
from ._stream_utils import stream_logs
from ._utils import (
    validate_managed_registry,
    get_validate_platform,
    get_custom_registry_credentials,
    get_yaml_template,
    prepare_source_location
)
from ._client_factory import cf_acr_registries_tasks

RUN_NOT_SUPPORTED = 'Run is only available for managed registries.'

logger = get_logger(__name__)


def acr_run(cmd,  # pylint: disable=too-many-locals
            client,
            registry_name,
            source_location,
            agent_pool_name=None,
            file=None,
            values=None,
            set_value=None,
            set_secret=None,
            cmd_value=None,
            no_format=False,
            no_logs=False,
            no_wait=False,
            timeout=None,
            resource_group_name=None,
            platform=None,
            auth_mode=None,
            log_template=None):

    _, resource_group_name = validate_managed_registry(
        cmd, registry_name, resource_group_name, RUN_NOT_SUPPORTED)

    if cmd_value and file:
        raise CLIError(
            "Azure Container Registry can run with either "
            "--cmd myCommand /dev/null or "
            "-f myFile mySourceLocation, but not both.")

    client_registries = cf_acr_registries_tasks(cmd.cli_ctx)
    source_location = prepare_source_location(
        cmd, source_location, client_registries, registry_name, resource_group_name)

    platform_os, platform_arch, platform_variant = get_validate_platform(cmd, platform)

    EncodedTaskRunRequest, FileTaskRunRequest, PlatformProperties = cmd.get_models(
        'EncodedTaskRunRequest', 'FileTaskRunRequest', 'PlatformProperties', operation_group='runs')

    if source_location:
        request = FileTaskRunRequest(
            task_file_path=file if file else ACR_TASK_YAML_DEFAULT_NAME,
            values_file_path=values,
            values=(set_value if set_value else []) + (set_secret if set_secret else []),
            source_location=source_location,
            timeout=timeout,
            platform=PlatformProperties(
                os=platform_os,
                architecture=platform_arch,
                variant=platform_variant
            ),
            credentials=get_custom_registry_credentials(
                cmd=cmd,
                auth_mode=auth_mode
            ),
            agent_pool_name=agent_pool_name,
            log_template=log_template
        )
    else:
        yaml_template = get_yaml_template(cmd_value, timeout, file)
        import base64
        request = EncodedTaskRunRequest(
            encoded_task_content=base64.b64encode(yaml_template.encode()).decode(),
            values=(set_value if set_value else []) + (set_secret if set_secret else []),
            source_location=source_location,
            timeout=timeout,
            platform=PlatformProperties(
                os=platform_os,
                architecture=platform_arch,
                variant=platform_variant
            ),
            credentials=get_custom_registry_credentials(
                cmd=cmd,
                auth_mode=auth_mode
            ),
            agent_pool_name=agent_pool_name,
            log_template=log_template
        )

    queued = LongRunningOperation(cmd.cli_ctx)(client_registries.begin_schedule_run(
        resource_group_name=resource_group_name,
        registry_name=registry_name,
        run_request=request))

    run_id = queued.run_id
    logger.warning("Queued a run with ID: %s", run_id)

    if no_wait:
        return queued

    logger.warning("Waiting for an agent...")

    if no_logs:
        from ._run_polling import get_run_with_polling
        return get_run_with_polling(cmd, client, run_id, registry_name, resource_group_name)

    return stream_logs(cmd, client, run_id, registry_name, resource_group_name, no_format, True)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=unused-argument
from azure.cli.core.util import sdk_no_wait
from azure.mgmt.synapse.models import WorkspaceAadAdminInfo

from .._client_factory import cf_synapse_client_workspace_factory
from ..util import get_tenant_id
from ..constant import AdministratorType


# Synapse SQL ad-admin
def create_workspace_sql_aad_admin(cmd, client, resource_group_name, workspace_name, login_name, object_id,
                                   no_wait=False):
    """
    Set a Workspace SQL AD admin.
    """
    workspace_client = cf_synapse_client_workspace_factory(cmd.cli_ctx)
    workspace_object = workspace_client.get(resource_group_name, workspace_name)

    workspace_id = workspace_object.id
    tenant_id = get_tenant_id()
    workspace_aad_admin_info = WorkspaceAadAdminInfo(id=workspace_id, login=login_name, sid=object_id,
                                                     administrator_type=AdministratorType, tenant_id=tenant_id)
    return sdk_no_wait(no_wait, client.begin_create_or_update, resource_group_name, workspace_name,
                       workspace_aad_admin_info)


def update_workspace_sql_aad_admin(instance, login_name=None, object_id=None):
    """
    Update a Workspace SQL AD admin.
    """
    instance.login = login_name or instance.login
    instance.sid = object_id or instance.sid
    return instance
